@@ -8,13 +8,48 @@ module can change at any time without warning.
 
 from __future__ import annotations
 
+import copy
+import warnings
 from typing import Any, Dict, List, Optional, Protocol, Tuple, Union, cast
 
+import numpy
 import shapely
+import shapely.affinity
 import shapely.geometry
-from shapely.geometry import MultiLineString, MultiPolygon, Polygon
+import shapely.validation
+from shapely.geometry import (
+    LinearRing,
+    LineString,
+    MultiLineString,
+    MultiPolygon,
+    Point,
+    Polygon,
+)
 
-Point = Tuple[float, float]
+XY = Tuple[float, float]
+
+
+class AntimeridianWarning(UserWarning):
+    """Base class for all package-specific warnings."""
+
+
+class FixWindingWarning(AntimeridianWarning):
+    """The input shape is wound clockwise (instead of counter-clockwise), so
+    this package is reversing the winding order before fixing the shape.
+    """
+
+    MESSAGE = (
+        "The exterior ring of this shape is wound "
+        "clockwise. Since this is a common error in real-world "
+        "geometries, this package is reversing the exterior coordinates of the "
+        "input shape before running its algorithm. If you know that your input "
+        "shape is correct (i.e. if your data encompasses both poles), pass "
+        "`fix_winding=False`."
+    )
+
+    @classmethod
+    def warn(cls) -> None:
+        warnings.warn(cls.MESSAGE, cls, stacklevel=2)
 
 
 class GeoInterface(Protocol):
@@ -35,13 +70,14 @@ def fix_geojson(
     *,
     force_north_pole: bool = False,
     force_south_pole: bool = False,
+    fix_winding: bool = True,
 ) -> Dict[str, Any]:
     """Fixes a GeoJSON object that crosses the antimeridian.
 
     If the object does not cross the antimeridian, it is returned unchanged.
 
-    See :py:func:`fix_polygon` for a description of the ``force_north_pole`` and
-    ``force_south_pole`` arguments.
+    See :py:func:`fix_polygon` for a description of the ``force_north_pole``,
+    ``force_south_pole``, and ``fix_winding`` arguments.
 
     Args:
         geojson: A GeoJSON object as a dictionary
@@ -49,6 +85,8 @@ def fix_geojson(
             joined segments to enclose the north pole.
         force_south_pole: If the polygon crosses the antimeridian, force the
             joined segments to enclose the south pole.
+        fix_winding: If the polygon is wound clockwise, reverse its
+            coordinates before applying the algorithm.
 
     Return:
         The same GeoJSON with a fixed geometry or geometries
@@ -64,6 +102,7 @@ def fix_geojson(
             geometry,
             force_north_pole=force_north_pole,
             force_south_pole=force_south_pole,
+            fix_winding=fix_winding,
         )
         return geojson
     elif type_ == "FeatureCollection":
@@ -75,6 +114,7 @@ def fix_geojson(
                 feature,
                 force_north_pole=force_north_pole,
                 force_south_pole=force_south_pole,
+                fix_winding=fix_winding,
             )
         geojson["features"] = features
         return geojson
@@ -83,6 +123,7 @@ def fix_geojson(
             geojson,
             force_north_pole=force_north_pole,
             force_south_pole=force_south_pole,
+            fix_winding=fix_winding,
         )
 
 
@@ -123,20 +164,23 @@ def fix_shape(
     *,
     force_north_pole: bool = False,
     force_south_pole: bool = False,
+    fix_winding: bool = True,
 ) -> Dict[str, Any]:
     """Fixes a shape that crosses the antimeridian.
 
-    See :py:func:`fix_polygon` for a description of the ``force_north_pole`` and
-    ``force_south_pole`` arguments.
+    See :py:func:`fix_polygon` for a description of the ``force_north_pole``,
+    ``force_south_pole``, and ``fix_winding`` arguments.
 
     Args:
-        shape: A polygon or multi-polygon, either as a dictionary or as a
-            :py:class:`GeoInterface`. Uses :py:func:`shapely.geometry.shape`
-            under the hood.
+        shape: A polygon, multi-polygon, line string, or multi-line string,
+            either as a dictionary or as a :py:class:`GeoInterface`. Uses
+            :py:func:`shapely.geometry.shape` under the hood.
         force_north_pole: If the polygon crosses the antimeridian, force the
             joined segments to enclose the north pole.
         force_south_pole: If the polygon crosses the antimeridian, force the
             joined segments to enclose the south pole.
+        fix_winding: If the polygon is wound clockwise, reverse its
+            coordinates before applying the algorithm.
 
     Returns:
         The fixed shape as a dictionary
@@ -150,6 +194,7 @@ def fix_shape(
                     geom,
                     force_north_pole=force_north_pole,
                     force_south_pole=force_south_pole,
+                    fix_winding=fix_winding,
                 )
             ),
         )
@@ -161,14 +206,21 @@ def fix_shape(
                     geom,
                     force_north_pole=force_north_pole,
                     force_south_pole=force_south_pole,
+                    fix_winding=fix_winding,
                 )
             ),
         )
+    elif geom.geom_type == "LineString":
+        return cast(Dict[str, Any], shapely.geometry.mapping(fix_line_string(geom)))
+    elif geom.geom_type == "MultiLineString":
+        return cast(
+            Dict[str, Any], shapely.geometry.mapping(fix_multi_line_string(geom))
+        )
     else:
         raise ValueError(f"unsupported geom_type: {geom.geom_type}")
 
 
-def segment_shape(shape: Dict[str, Any] | GeoInterface) -> List[List[Point]]:
+def segment_shape(shape: Dict[str, Any] | GeoInterface) -> List[List[XY]]:
     geom = shapely.geometry.shape(shape)
     if geom.geom_type == "Polygon":
         return segment_polygon(geom)
@@ -186,11 +238,12 @@ def fix_multi_polygon(
     *,
     force_north_pole: bool = False,
     force_south_pole: bool = False,
+    fix_winding: bool = True,
 ) -> MultiPolygon:
     """Fixes a :py:class:`shapely.geometry.MultiPolygon`.
 
-    See :py:func:`fix_polygon` for a description of the ``force_north_pole`` and
-    ``force_south_pole`` arguments.
+    See :py:func:`fix_polygon` for a description of the ``force_north_pole``,
+    ``force_south_pole``, and ``fix_winding`` arguments.
 
     Args:
         multi_polygon: The multi-polygon
@@ -198,6 +251,8 @@ def fix_multi_polygon(
             joined segments to enclose the north pole.
         force_south_pole: If the polygon crosses the antimeridian, force the
             joined segments to enclose the south pole.
+        fix_winding: If the polygon is wound clockwise, reverse its
+            coordinates before applying the algorithm.
 
     Returns:
         The fixed multi-polygon
@@ -208,18 +263,24 @@ def fix_multi_polygon(
             polygon,
             force_north_pole=force_north_pole,
             force_south_pole=force_south_pole,
+            fix_winding=fix_winding,
         )
     return MultiPolygon(polygons)
 
 
 def fix_polygon(
-    polygon: Polygon, *, force_north_pole: bool = False, force_south_pole: bool = False
+    polygon: Polygon,
+    *,
+    force_north_pole: bool = False,
+    force_south_pole: bool = False,
+    fix_winding: bool = True,
 ) -> Union[Polygon, MultiPolygon]:
     """Fixes a :py:class:`shapely.geometry.Polygon`.
 
-    If the input polygon is a single polygon that is wound clockwise and doesn't
-    cross the antimeridian, it will be corrected by adding a counter-clockwise
-    polygon from (-180, -90) to (180, 90) as its exterior.
+    If the input polygon is wound clockwise, it will be fixed to be wound
+    counterclockwise _unless_ ``fix_winding`` is ``False``, in which case it
+    will be corrected by adding a counterclockwise polygon from (-180, -90) to
+    (180, 90) as its exterior.
 
     In rare cases, the underlying algorithm might need a little help to fix the polygon.
     For example, a polygon that just barely crosses over a pole might have very
@@ -228,19 +289,29 @@ def fix_polygon(
     and ``force_south_pole``, for those cases. Most users can ignore these
     flags.
 
+    If either ``force_north_pole`` or ``force_south_pole`` is ``True``,
+    ``fix_winding`` is set to ``False``.
+
     Args:
         polygon: The input polygon
         force_north_pole: If the polygon crosses the antimeridian, force the
             joined segments to enclose the north pole.
         force_south_pole: If the polygon crosses the antimeridian, force the
             joined segments to enclose the south pole.
+        fix_winding: If the polygon is wound clockwise, reverse its
+            coordinates before applying the algorithm.
 
     Returns:
         The fixed polygon, either as a single polygon or a multi-polygon (if it
         was split)
     """
+    if force_north_pole or force_south_pole:
+        fix_winding = False
     polygons = fix_polygon_to_list(
-        polygon, force_north_pole=force_north_pole, force_south_pole=force_south_pole
+        polygon,
+        force_north_pole=force_north_pole,
+        force_south_pole=force_south_pole,
+        fix_winding=fix_winding,
     )
     if len(polygons) == 1:
         polygon = polygons[0]
@@ -255,7 +326,43 @@ def fix_polygon(
         return MultiPolygon(polygons)
 
 
-def segment_polygon(polygon: Polygon) -> List[List[Point]]:
+def fix_line_string(line_string: LineString) -> Union[LineString, MultiLineString]:
+    """Fixes a :py:class:`shapely.geometry.LineString`.
+
+    Args:
+        line_string: The input line string
+
+    Returns:
+        The fixed line string, either as a single line string or a multi-line
+        string (if it was split)
+    """
+    segments = segment(list(line_string.coords))
+    if not segments:
+        return line_string
+    else:
+        return MultiLineString(segments)
+
+
+def fix_multi_line_string(multi_line_string: MultiLineString) -> MultiLineString:
+    """Fixes a :py:class:`shapely.geometry.MultiLineString`.
+
+    Args:
+        multi_line_string: The input multi line string
+
+    Returns:
+        The fixed multi line string
+    """
+    line_strings = list()
+    for line_string in multi_line_string.geoms:
+        fixed = fix_line_string(line_string)
+        if isinstance(fixed, LineString):
+            line_strings.append(fixed)
+        else:
+            line_strings.extend(fixed.geoms)
+    return MultiLineString(line_strings)
+
+
+def segment_polygon(polygon: Polygon) -> List[List[XY]]:
     segments = segment(list(polygon.exterior.coords))
     if not segments:
         segments = [list(polygon.exterior.coords)]
@@ -269,21 +376,42 @@ def segment_polygon(polygon: Polygon) -> List[List[Point]]:
 
 
 def fix_polygon_to_list(
-    polygon: Polygon, *, force_north_pole: bool, force_south_pole: bool
+    polygon: Polygon,
+    *,
+    force_north_pole: bool,
+    force_south_pole: bool,
+    fix_winding: bool,
 ) -> List[Polygon]:
     segments = segment(list(polygon.exterior.coords))
     if not segments:
-        return [polygon]
+        if fix_winding and (
+            not shapely.is_ccw(polygon.exterior)
+            or any(shapely.is_ccw(interior) for interior in polygon.interiors)
+        ):
+            FixWindingWarning.warn()
+            return [shapely.geometry.polygon.orient(polygon)]
+        else:
+            return [polygon]
     else:
         interiors = []
         for interior in polygon.interiors:
             interior_segments = segment(list(interior.coords))
             if interior_segments:
+                if fix_winding:
+                    unwrapped_linearring = LinearRing(
+                        list((x % 360, y) for x, y in interior.coords)
+                    )
+                    if shapely.is_ccw(unwrapped_linearring):
+                        FixWindingWarning.warn()
+                        interior_segments = segment(list(reversed(interior.coords)))
                 segments.extend(interior_segments)
             else:
                 interiors.append(interior)
     segments = extend_over_poles(
-        segments, force_north_pole=force_north_pole, force_south_pole=force_south_pole
+        segments,
+        force_north_pole=force_north_pole,
+        force_south_pole=force_south_pole,
+        fix_winding=fix_winding,
     )
     polygons = build_polygons(segments)
     assert polygons
@@ -298,12 +426,17 @@ def fix_polygon_to_list(
     return polygons
 
 
-def segment(coords: List[Point]) -> List[List[Point]]:
+def segment(coords: List[XY]) -> List[List[XY]]:
     segment = []
     segments = []
     for i, point in enumerate(coords):
-        # Ensure all longitudes are between -180 and 180
-        if point[0] != 180:
+        # Ensure all longitudes are between -180 and 180, and that tiny floating
+        # point differences are ignored.
+        if numpy.isclose(point[0], 180):
+            coords[i] = (180, point[1])
+        elif numpy.isclose(point[0], -180):
+            coords[i] = (-180, point[1])
+        else:
             coords[i] = (((point[0] + 180) % 360) - 180, point[1])
     for start, end in zip(coords, coords[1:]):
         segment.append(start)
@@ -323,10 +456,17 @@ def segment(coords: List[Point]) -> List[List[Point]]:
     elif coords[-1] == segments[0][0]:
         # Join polygons
         segments[0] = segment + segments[0]
+    else:
+        segment.append(coords[-1])
+        segments.append(segment)
     return segments
 
 
-def crossing_latitude(start: Point, end: Point) -> float:
+def crossing_latitude(start: XY, end: XY) -> float:
+    if abs(start[0]) == 180:
+        return start[1]
+    elif abs(end[0]) == 180:
+        return end[1]
     latitude_delta = end[1] - start[1]
     if end[0] > 0:
         return round(
@@ -343,11 +483,12 @@ def crossing_latitude(start: Point, end: Point) -> float:
 
 
 def extend_over_poles(
-    segments: List[List[Point]],
+    segments: List[List[XY]],
     *,
     force_north_pole: bool,
     force_south_pole: bool,
-) -> List[List[Point]]:
+    fix_winding: bool,
+) -> List[List[XY]]:
     left_starts = list()
     right_starts = list()
     left_ends = list()
@@ -365,21 +506,41 @@ def extend_over_poles(
     left_starts.sort(key=lambda v: v[1])
     right_ends.sort(key=lambda v: v[1], reverse=True)
     right_starts.sort(key=lambda v: v[1], reverse=True)
+    is_over_north_pole = False
+    is_over_south_pole = False
+    original_segments = copy.deepcopy(segments)
     # If there's no segment ends between a start and the pole, extend the
     # segment over the pole.
     if left_ends and (
         force_south_pole or not left_starts or left_ends[0][1] < left_starts[0][1]
     ):
+        is_over_south_pole = True
         segments[left_ends[0][0]] += [(-180, -90), (180, -90)]
     if right_ends and (
         force_north_pole or not right_starts or right_ends[0][1] > right_starts[0][1]
     ):
+        is_over_north_pole = True
         segments[right_ends[0][0]] += [(180, 90), (-180, 90)]
-    return segments
+    if fix_winding and is_over_north_pole and is_over_south_pole:
+        # These assertions are here because we're assuming that we set
+        # `fix_winding` to `False` up in `fix_polygon` if either of the
+        # `force_*` variables are set.
+        assert not force_north_pole
+        assert not force_south_pole
+
+        # If we're over both poles and we haven't explicitly disabled the
+        # fix behavior, reverse all segments, effectively reversing the
+        # winding order.
+        FixWindingWarning.warn()
+        for segment in original_segments:
+            segment.reverse()
+        return original_segments
+    else:
+        return segments
 
 
 def build_polygons(
-    segments: List[List[Point]],
+    segments: List[List[XY]],
 ) -> List[Polygon]:
     if not segments:
         return []
@@ -425,11 +586,16 @@ def build_polygons(
         # This segment should self-joining, so just build the rest of the
         # polygons without it.
         polygons = build_polygons(segments)
-        polygons.append(Polygon(segment))
+        if not all(p == segment[0] for p in segment):
+            # If every point is the same, then we don't need it in the output
+            # set of polygons. This happens if, e.g., one corner of an input
+            # polygon is on the antimeridian.
+            # https://github.com/gadomski/antimeridian/issues/45#issuecomment-1614586166
+            polygons.append(Polygon(segment))
         return polygons
 
 
-def is_self_closing(segment: List[Point]) -> bool:
+def is_self_closing(segment: List[XY]) -> bool:
     is_right = segment[-1][0] == 180
     return segment[0][0] == segment[-1][0] and (
         (is_right and segment[0][1] > segment[-1][1])
@@ -481,6 +647,43 @@ def bbox(shape: Dict[str, Any] | GeoInterface) -> List[float]:
         )
 
 
+def centroid(shape: Dict[str, Any] | GeoInterface) -> Point:
+    """Calculates the centroid for a polygon or multipolygon.
+
+    Polygons are easy, we just use :py:func:`shapely.centroid`. For
+    multi-polygons, the antimeridian is taken into account by calculating the
+    centroid from an identical multi-polygon with coordinates in [0, 360).
+
+    Args:
+        shape: The polygon or multipolygon for which to calculate the centroid.
+
+    Returns:
+        Point: The centroid.
+    """
+    # Inspired by
+    # https://github.com/stactools-packages/sentinel2/blob/f90f5fa006459e9bb59bfd327d9199e5259ec4a7/src/stactools/sentinel2/stac.py#L192-L208
+    geom = shapely.geometry.shape(shape)
+    if geom.geom_type == "Polygon":
+        return cast(Point, geom.centroid)
+    elif geom.geom_type == "MultiPolygon":
+        geoms = list()
+        for component in geom.geoms:
+            if any(c[0] < 0 for c in component.exterior.coords):
+                geoms.append(shapely.affinity.translate(component, xoff=+360))
+            else:
+                geoms.append(component)
+        centroid = cast(
+            Point, shapely.validation.make_valid(MultiPolygon(geoms)).centroid
+        )
+        if centroid.x > 180:
+            centroid = Point(centroid.x - 360, centroid.y)
+        return centroid
+    else:
+        raise ValueError(
+            f"unsupported geom_type for centroid calculation: {geom.geom_type}"
+        )
+
+
 def is_coincident_to_antimeridian(polygon: Polygon) -> bool:
     for start, end in zip(polygon.exterior.coords, polygon.exterior.coords[1:]):
         if abs(start[0]) == 180 and start[0] == end[0]:
